/**
 * NORRIS.js
 * Copyright (c) 2011 Ruben L Z Tan
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
 * IN THE SOFTWARE.
 */
'use strict';

var fs      = require('fs');
var path    = require('path');
var Emitter = require('events').EventEmitter;
var exec    = require('child_process').exec;
var vows    = require('vows');
var assert  = require('assert');
var mocker  = require(__dirname + '/../test/mock/base.mock');

var Test = module.exports = {
    // Properties
    types   : ['all', 'unit', 'hook', 'full'],
    options : ['p', 'pass-only', 'f', 'fail-only', 'v', 'verbose', 'e', 'exception', 'h', 'help'],
    
    // Methods
    run     : run
};

/**----------------------------------------------------------------------------
 * Runs the command
 */
function run(util, commands, options) {
    var self = this;
    var type = commands[0];
    
    // Check if the test that we want to run is a supported type or just a 
    // list of unit test files
    if (!type || -1 !== self.types.indexOf(type)) {
        type = (null == type) ? 'all' : type;
    }
    // Not a supported type. Treat it as a namespace to a file 
    else {
        type = [];
        for (var i = commands.length; i--;) {
            var cmd = commands[i];
            type[cmd.replace(/\.g/g, '/') + '.test.js'] = true;
        }
    }
    
    // Check if there are unsupported options
    var unsupported = util.options(self.options).input(options).check();
    if (unsupported.length > 0) {
        util.color('#red[Unsupported options %s\n]', unsupported.join(', '));
        util.help('test');
        return;
    }
    
    // Options
    var passOnly    = util.has('p', 'pass-only');
    var failOnly    = util.has('f', 'fail-only');
    var exception   = util.has('e', 'exception');
    var verbose     = util.has('v', 'verbose');
    var help        = util.has('h', 'help');
    
    if (help) {
        util.help('test');
        return;
    }
    
    var promise = new Emitter();
    
    var reporter = {};
    
    (function initReporter(reporter) {
        var json = {};
        reporter.results = [];
        reporter.reset = function() {}
        reporter.report = function(obj) { this.results.push(obj); }
        reporter.print = function(str) {}
    })(reporter);
    
    // Handle errors when running the test cases
    promise.on('error', function onErr(err) {
        util.log('Unexpected error encountered');
        util.log(err.stack);
    });
    
    // Handle file event
    promise.on('file', function onFile(type, path, name) {
        var test = null;
            
        // Utility function to auto-append whitespaces
        function conformTo(m, size) {
            m = m + '';
            while (m.length < size) {
                m = ' ' + m;
            }
            return m;
        }
        
        function checkDone(util, reporter) {
            util.totalFiles--;
            if (util.totalFiles <= 0)
                promise.emit('done', reporter.results);
        }
        
        try {
            test = require(path).make(vows, assert, mocker.load(name));
            
            test.run({ reporter : reporter }, function onRun(result) {
                var pass    = conformTo(result.honored, 4);
                var fail    = conformTo(result.broken, 4);
                var err     = conformTo(result.errored, 4);
                var total   = conformTo(result.total, 4);
                var time    = result.time + 'ms';
                var color   = (result.broken > 0 || result.errored > 0) ? 'red' : 'cyan';
                
                util.color('#bold[#' + color + '[' + name + ']]', false);
                
                // Calculate how many whitespaces to use
                var wsl = 30 - name.length;
                var ws  = '';
                for (var i = wsl; i--;) ws += ' ';
                util.color(ws, false);
                util.color('#green[Pass: ' + pass + ']\t#yellow[Fail: ' + fail + ']\t', false);
                util.color('#red[Erro: ' + err + ']\t#blue[Total: ' + total + ']\t#magenta[Time: ' + time + ']');
                
                // Check if we've run all the test cases
                checkDone(util, reporter);
            });
        }
        catch(err) {
            checkDone(util, reporter);
        }
    });
    
    // Handle completion of test cases
    promise.on('done', function onDone(results) {
        if (verbose) {
            passOnly    = true;
            failOnly    = true;
            exception   = true;
        }
        
        if (passOnly || failOnly) {
            util.log('\nTest Results: ');
            verboseDetails(util, results, {
                pass        : passOnly,
                fail        : failOnly,
                exception   : exception,
                verbose     : verbose
            });
        }
    });
    
    // Run the test suites!
    if ('string' === typeof type) {
        promise = bulkRun(util, promise, type);
    }
    else {
        promise = selectiveRun(util, promise, type);
    }
}

/**----------------------------------------------------------------------------
 * Runs all tests of a specific type
 */
function bulkRun(util, promise, type) {
    if ('all' === type)
        type = ['unit', 'hook', 'full'];
    else
        type = [type];
        
    util.totalFiles = 0;
        
    function walk(basepath) {
        fs.readdir(basepath, function onReadDir(err, files) {
            if (err) {
                util.log('Attempted to load ' + basepath);
                promise.emit('error', new Error(err));
            }
            else {
                // Check if it's a directory or a file
                for (var i = 0, len = files.length; i < len; i++) {
                    var name = files[i];
                    var fullPath = basepath + '/' + name;
                    
                    var stat = fs.statSync(fullPath);
                    if (stat.isDirectory()) {
                        walk(fullPath);
                    }
                    else {
                        util.totalFiles++;
                        promise.emit('file', 'bulk', fullPath, name);
                    }
                }
            }
        });
    }
    
    for (var i = type.length; i--;) {
        walk(path.normalize(__dirname + '/../test/' + type[i]));
    }
    
    return promise;
}

/**----------------------------------------------------------------------------
 * Selectively run the test cases
 */
function selectiveRun(util, promise, files) {
    var base = path.normalize(__dirname + '/../test');
    
    for (var i in files) {
        var file = i;
        var fullPath = base + '/' + file;
        
        // Only process if the file exists
        if (path.existsSync(fullPath)) {
            // Get the filename
            var name = path.basename(file, path.extname(file));
            promise.emit('file', 'file', fullPath, name);
        }
        else {
            promise.emit('error', new Error(fullPath + ' not found'));
        }
    }
    
    return promise;
}

/**----------------------------------------------------------------------------
 * Verbose output of the test results
 */
function verboseDetails(util, results, options) {
    var suites  = [];
    var curr    = 0;
    
    for (var i in results) {
        var result  = results[i];
        var type    = result.shift();
        
        if ('subject' === type) {
            suites[curr] = { title : result.shift(), status : true };
            continue;
        } else if ('finished' === type) {
            curr++;
            continue;
        }
                
        var data = result.shift();
        
        switch (type) {
            case 'context' :
                if (undefined === suites[curr].context)
                    suites[curr].context = [];
                    
                suites[curr].context[data] = { title : data, status : true, vows : [] };
                break;
                
            case 'vow' :
                suites[curr].context[data.context].vows.push(data);
                if ('honored' !== data.status) {
                    var suite = suites[curr];
                    suite.status                        = false;
                    suite.context[data.context].status  = false;
                    suites[curr]                        = suite;
                }
                break;
            
            default : continue;
        }
    }
    
    // Now that we have all the contexts buried in the right subjects, time to display them!
    for (var i in suites) {
        var suite   = suites[i];
        var title   = suite.title;
        var status  = suite.status;
        var color   = (status) ? 'cyan' : 'red';
        
        if (options.pass && status || options.fail && !status)
            util.color('- #' + color + '[' + title + ']');
            
        for (var ctxTitle in suite.context) {
            var context     = suite.context[ctxTitle];
            var ctxStatus   = context.status;
            var ctxColor    = (ctxStatus) ? 'magenta' : 'red';
            
            if (options.pass && ctxStatus || options.fail && !ctxStatus)
                util.color('  • #' + ctxColor + '[' + ctxTitle + ']');
                
            for (var j in context.vows) {
                var vow         = context.vows[j];
                var vowTitle    = vow.title;
                var vowStatus   = ('honored' === vow.status);
                var vowColor    = (vowStatus) ? 'green' : 'red';
                var vowIcon     = (vowStatus) ? '√' : '✗';
                var vowError    = vow.exception;
                
                if (options.pass && vowStatus || options.fail && !vowStatus) {
                    util.color('    #' + vowColor + '[' + vowIcon + ' ' + vowTitle + ']');
                }
                
                if (options.exception && !vowStatus && (!options.pass || options.verbose)) {
                    util.color('      #blue[' + vowError.replace(/\n/g, '\n      ') + ']');
                }
            }
        }
    }
    
}
